"""Represents a ShutIt background command.

    - options to:
        - cancel command
        - get status (running, suspended etc)
        - check_timeout

     state     The state is given by a sequence of characters, for example, ``RWNA''.  The first character indicates the run state of the process:

               I       Marks a process that is idle (sleeping for longer than about 20 seconds).
               R       Marks a runnable process.
               S       Marks a process that is sleeping for less than about 20 seconds.
               T       Marks a stopped process.
               U       Marks a process in uninterruptible wait.
               Z       Marks a dead process (a ``zombie'').


               Additional characters after these, if any, indicate additional state information:

               +       The process is in the foreground process group of its control terminal.
               <       The process has raised CPU scheduling priority.
               >       The process has specified a soft limit on memory requirements and is currently exceeding that limit; such a process is (necessarily) not swapped.
               A       the process has asked for random page replacement (VA_ANOM, from vadvise(2), for example, lisp(1) in a garbage collect).
               E       The process is trying to exit.
               L       The process has pages locked in core (for example, for raw I/O).
               N       The process has reduced CPU scheduling priority (see setpriority(2)).
               S       The process has asked for FIFO page replacement (VA_SEQL, from vadvise(2), for example, a large image processing program using virtual memory to sequentially
                       address voluminous data).
               s       The process is a session leader.
               V       The process is suspended during a vfork(2).
               W       The process is swapped out.
               X       The process is being traced or debugged.

ps -o stat= | sed 's/^\(.\)\(.*\)/\1/'
"""

import time
import logging


class ShutItBackgroundCommand(object):
	"""Background command in ShutIt
	"""
	def __init__(self,
	             sendspec,
	             block_other_commands=True,
	             background=True):
		# Stub this with a simple command for now
		self.sendspec             = sendspec
		self.block_other_commands = block_other_commands
		self.pid                  = None
		self.return_value         = None
		self.start_time           = None
		self.run_state            = 'N' # State as per ps man page, but 'C' == Complete, 'N' == not started
		self.cwd                  = self.sendspec.shutit_pexpect_child.send_and_get_output(' command pwd')


	def __str__(self):
		string = ''
		string += str(self.sendspec)
		string += '\nblock_other_commands: ' + str(self.block_other_commands)
		string += '\npid: ' + str(self.pid)
		string += '\nreturn_value: ' + str(self.return_value)
		string += '\nstart_time: ' + str(self.start_time)
		string += '\nrun_state: ' + str(self.run_state)
		string += '\ncwd: ' + str(self.cwd)
		return string


	def run_background_command(self):
		shutit_pexpect_child = self.sendspec.shutit_pexpect_child
		# Go to cwd
		shutit_pexpect_child.quick_send(' command cd ' + self.cwd)
		# record start time
		self.start_time = time.localtime()
		# run command
		if self.sendspec.check_exit:
			# Override - maybe log this TODO
			self.sendspec.check_exit = False
		#print 'sending: ' + self.sendspec.send
		#print 'in: ' + str(shutit_pexpect_child)
		shutit_pexpect_child.quick_send(self.sendspec.send)
		self.sendspec.started = True
		# Required to reset terminal after a background send.
		shutit_pexpect_child.reset_terminal()
		# record pid
		self.pid = shutit_pexpect_child.send_and_get_output(" echo ${!}")
		#print self.pid
		return True


	def check_background_command_state(self):
		#print self
		shutit_pexpect_child = self.sendspec.shutit_pexpect_child
		#print('self.sendspec.send: ' + str(self.sendspec.send))
		#print('self.sendspec.started: ' + str(self.sendspec.started))
		#print('self.run_state: ' + str(self.run_state))
		#print('self.pid: ' + str(self.pid))
		#print('self.return_value: ' + str(self.return_value))
		# Check the command has been started
		if not self.sendspec.started:
			#print 'not started?'
			return self.run_state
		self.run_state = shutit_pexpect_child.send_and_get_output(""" command ps -o stat """ + self.pid + """ | command sed '1d' """)
		# If the job is complete, collect the return value
		if self.run_state == '':
			self.run_state = 'C'
			# Stop this from blocking other commands from here.
			self.block_other_commands = False
		if isinstance(self.run_state,str) and self.run_state == 'C' and self.return_value is None:
			shutit_pexpect_child.quick_send(' wait ' + self.pid)
			self.return_value = shutit_pexpect_child.send_and_get_output(' echo $?')
		if isinstance(self.run_state,str) and self.run_state == 'C' and self.return_value is not None:
			# This job is complete, nothing to do.
			pass
		# TODO: honour sendspec.timeout
		#print 'returning: ' + self.run_state
		return self.run_state




"""Contains all the core ShutIt methods and functionality, and public interface
off to internal objects such as shutit_pexpect.
"""

# The MIT License (MIT)
#
# Copyright (C) 2014 OpenBet Limited
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# ITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import os
import socket
import time
import re
import getpass
import codecs
import datetime
import logging
import tarfile
import pexpect
import shutit_util
import shutit_setup
from shutit_module import ShutItFailException
from shutit_class import ShutIt


class ShutItGlobal(object):
	"""Single object to store all the separate ShutIt sessions.
	"""

	only_one = None
	report_final_messages = ''
	def __init__(self):
		"""Constructor.
		"""
		self.shutit_objects = []
		# Primitive singleton enforcer.
		assert self.only_one is None
		self.only_one         = True

		self.secret_words_set = set()
		self.logfile          = None
		self.loglevel         = None
		self.shutit_signal_id = None
		self.username         = os.environ.get('LOGNAME', '')
		# Environments are kept globally, as different sessions may re-connect to them.                                                                                                                 
		self.shutit_pexpect_session_environments = set()
		if self.username == '':
			try:
				if os.getlogin() != '':
					self.username = os.getlogin()
			except Exception:
				self.username = getpass.getuser()
			if self.username == '':
				shutit_util.handle_exit(msg='LOGNAME not set in the environment, ' + 'and login unavailable in python; ' + 'please set to your username.', exit_code=1)
		self.real_user        = os.environ.get('SUDO_USER', self.username)
		self.build_id         = (socket.gethostname() + '_' + self.real_user + '_' + str(time.time()) + '.' + str(datetime.datetime.now().microsecond))
		self.shutit_state_dir_base  = '/tmp/shutit_' + self.username
		self.shutit_state_dir       = self.shutit_state_dir_base + '/' + self.build_id
		self.shutit_state_dir_build_db_dir = self.shutit_state_dir + '/build_db'


	def add_shutit_session(self, shutit):
		self.shutit_objects.append(shutit)

	def create_session(self, session_type='bash', docker_image=None, rm=None):
		assert isinstance(session_type, str)
		new_shutit = ShutIt()
		self.add_shutit_session(new_shutit)
		# TODO: only makes sense in session that's already bash - check this
		if session_type == 'bash':
			shutit_util.parse_args(new_shutit)
			shutit_util.load_configs(shutit=new_shutit)
			shutit_setup.setup_host_child_environment(new_shutit)
			return new_shutit
		elif session_type == 'docker':
			shutit_util.parse_args(new_shutit)
			# Set the configuration up appropriately using overrides.
			if docker_image:
				new_shutit.build['config_overrides'].append(['build','base_image',docker_image])
			if rm:
				new_shutit.target['rm'] = True
			# Now 'load' the configs
			shutit_util.load_configs(new_shutit)
			target_child = shutit_setup.conn_docker_start_container(new_shutit,'target_child')
			shutit_setup.setup_host_child_environment(new_shutit)
			shutit_setup.setup_target_child_environment(new_shutit, target_child)
			return new_shutit
		else:
			new_shutit.fail('unhandled session type: ' + session_type)


	def do_final_messages(self):
		# Show final report messages (ie messages to show after standard report).
		if self.report_final_messages != '':
			self.shutit_objects[0].log(shutit_util.colourise(31,'\r\n\r\n' + self.report_final_messages + '\r\n\r\n'), level=logging.INFO, transient=True)


	def log(self, msg, add_final_message=False, level=logging.INFO, transient=False, newline=True, mask_password=True):
		"""Logging function.

		@param add_final_message: Add this log line to the final message output to the user
		@param level:             Python log level
		@param transient:         Just write to terminal, no new line. If not a
		                          terminal, write nothing.
		"""
		if mask_password:
			for password in shutit_global_object.secret_words_set:
				if password in msg:
					msg.replace(password,'REDACTED')
		if transient:
			if sys.stdout.isatty():
				if newline:
					msg += '\r\n'
				sys.stdout.write(msg)
			else:
				return True
		else:
			logging.log(level,msg)
			if add_final_message:
				self.report_final_messages = self.report_final_messages + '\r\n' + msg + '\r\n'
		return True

shutit_global_object = ShutItGlobal()
shutit_global_object.add_shutit_session(ShutIt())



"""Represents a ShutItLoginStack object.

Every time ShutItPexpectSession.login() is run a new item is added, and every
time it a corresponding logout() function is called, it is popped.

It also holds within it information about ShutItBackgroundCommand objects
belonging to this login.
"""

import shutit_global
import logging
from shutit_background import ShutItBackgroundCommand

class ShutItLoginStack(object):

	def __init__(self):
		"""
		"""
		self.stack = []

	def append(self, login_id):
		self.stack.append(ShutItLoginStackItem(login_id))
		return True

	def pop(self):
		return self.stack.pop()

	def length(self):
		return len(self.stack)

	def get_current_login_id(self):
		if len(self.stack):
			return self.stack[-1].login_id
		else:
			return None

	def get_current_login_item(self):
		if len(self.stack):
			return self.stack[-1]
		else:
			return None

	def find_sendspec(self):
		for stack_item in self.stack:
			if stack_item.find_sendspec():
				return True
		return False

	def __str__(self):
		string = ''
		string += '\nstack: ' + str(self.stack)
		for item in self.stack:
			string += '\nlogin_item: ' + str(item)
		return string



class ShutItLoginStackItem(object):

	def __init__(self,login_id):
		"""
		"""
		self.login_id           = login_id
		self.background_objects = []

	def append_background_send(self,sendspec):
		shutit_background_command_object = ShutItBackgroundCommand(sendspec)
		self.background_objects.append(shutit_background_command_object)
		return shutit_background_command_object

	def has_blocking_background_send(self):
		for background_object in self.background_objects:
			if background_object.block_other_commands:
				shutit_global.shutit_global_object.log('The blocking send object is: ' + str(background_object),level=logging.DEBUG)
				return True

	def check_background_commands(self):
		unstarted_command_exists  = False
		incomplete_command_exists = False
		for background_object in self.background_objects:
			state = background_object.check_background_command_state()
			if state != 'C':
				if state == 'N':
					unstarted_command_exists = True
				else:
					incomplete_command_exists = True
		if incomplete_command_exists:
			# Started
			return False
		if unstarted_command_exists:
			# Start up an unstarted one (in order), and return False
			for background_object in self.background_objects:
				state = background_object.check_background_command_state()
				if state == 'N':
					background_object.run_background_command()
					return False
		# Nothing left to do - return True.
		return True

	def find_sendspec(self,sendspec):
		for background_object in self.background_objects:
			if background_object == sendspec:
				return True
		return False

	def __str__(self):
		string = ''
		string += '\nlogin_id: ' + str(self.login_id)
		for background_object in self.background_objects:
			string += '\nbackground_objects: ' + str(background_object)
		return string
